# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
from .. import constants


def molar_volume_from_unit_cell_volume(unit_cell_v, z):
    """
    Converts a unit cell volume from Angstroms^3 per unitcell,
    to m^3/mol.

    :param unit_cell_v: Unit cell volumes [A^3/unit cell].
    :type unit_cell_v: float

    :param z: Number of formula units per unit cell.
    :type z: float

    :returns: Volume [m^3/mol]
    :rtype: float
    """
    V = unit_cell_v * constants.Avogadro / 1.0e30 / z
    return V


def cell_parameters_to_vectors(cell_parameters):
    """
    Converts cell parameters to unit cell vectors.

    :param cell_parameters: An array containing the three lengths of the
        unit cell vectors [m], and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.
    :type cell_parameters: numpy.array (1D)

    :returns: The three vectors defining the parallelopiped cell [m].
        This function assumes that the first cell vector is colinear with the
        x-axis, and the second is perpendicular to the z-axis, and the third is
        defined in a right-handed sense.
    :rtype: numpy.array (2D)
    """
    a, b, c, alpha_deg, beta_deg, gamma_deg = cell_parameters
    alpha = np.radians(alpha_deg)
    beta = np.radians(beta_deg)
    gamma = np.radians(gamma_deg)

    n2 = (np.cos(alpha) - np.cos(gamma) * np.cos(beta)) / np.sin(gamma)
    M = np.array(
        [
            [a, 0, 0],
            [b * np.cos(gamma), b * np.sin(gamma), 0],
            [c * np.cos(beta), c * n2, c * np.sqrt(np.sin(beta) ** 2 - n2**2)],
        ]
    )
    return M


def cell_vectors_to_parameters(M):
    """
    Converts unit cell vectors to cell parameters.

    :param M: The three vectors defining the parallelopiped cell [m].
        This function assumes that the first cell vector is colinear with the
        x-axis, the second is perpendicular to the z-axis, and the third is
        defined in a right-handed sense.
    :type M: numpy.array (2D)


    :returns: An array containing the three lengths of the unit cell vectors [m],
        and the three angles [degrees].
        The first angle (:math:`\\alpha`) corresponds to the angle between the
        second and the third cell vectors, the second (:math:`\\beta`) to the
        angle between the first and third cell vectors, and the third
        (:math:`\\gamma`) to the angle between the first and second vectors.
    :rtype: numpy.array (1D)
    """

    assert M[0, 1] == 0
    assert M[0, 2] == 0
    assert M[1, 2] == 0

    a = M[0, 0]
    b = np.sqrt(np.power(M[1, 0], 2.0) + np.power(M[1, 1], 2.0))
    c = np.sqrt(
        np.power(M[2, 0], 2.0) + np.power(M[2, 1], 2.0) + np.power(M[2, 2], 2.0)
    )

    gamma = np.arccos(M[1, 0] / b)
    beta = np.arccos(M[2, 0] / c)
    alpha = np.arccos(M[2, 1] / c * np.sin(gamma) + np.cos(gamma) * np.cos(beta))

    gamma_deg = np.degrees(gamma)
    beta_deg = np.degrees(beta)
    alpha_deg = np.degrees(alpha)

    return np.array([a, b, c, alpha_deg, beta_deg, gamma_deg])



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import matplotlib.pyplot as plt


def pretty_plot():
    """
    Makes pretty plots. Overwrites the matplotlib default settings to allow
    for better fonts. Slows down plotting
    """
    import matplotlib.pyplot as plt

    plt.rc("text", usetex=True)
    plt.rcParams["text.latex.preamble"] = "\\usepackage{relsize}"
    plt.rc("font", family="sanserif")


def plot_projected_elastic_properties(
    mineral, plot_types, axes, n_zenith=31, n_azimuth=91, n_divs=100
):
    """
    :param mineral: Mineral object on which calculations should be done
    :type mineral: :class:`burnman.Mineral`

    :param plot_types: Plot types must be one of the following
        * 'vp' - V$_{P}$ (km/s)
        * 'vs1' - 'V$_{S1}$ (km/s)
        * 'vs2' - V$_{S2}$ (km/s)
        * 'vp/vs1' - V$_{P}$/V$_{S1}$
        * 'vp/vs2' - V$_{P}$/V$_{S2}$
        * 's anisotropy' - S-wave anisotropy (%), 200(vs1s - vs2s)/(vs1s + vs2s))
        * 'linear compressibility' - Linear compressibility (GPa$^{-1}$)
        * 'youngs modulus' - Youngs Modulus (GPa)
    :type plot_types: list of str

    :param axes: axes objects to be modified.
        Must be initialised with projection='polar'.
    :type axes: matplotlib.pyplot.axes objects

    :param n_zenith: Number of zeniths (plot resolution).
    :type n_zenith: int

    :param n_azimuth: Number of azimuths (plot resolution).
    :type n_azimuth: int

    :param n_divs: Number of divisions for the color scale.
    :type n_divs: int
    """

    assert len(plot_types) == len(axes)

    zeniths = np.linspace(np.pi / 2.0, np.pi, n_zenith)
    azimuths = np.linspace(0.0, 2.0 * np.pi, n_azimuth)
    Rs = np.sin(zeniths) / (1.0 - np.cos(zeniths))
    r, theta = np.meshgrid(Rs, azimuths)

    vps = np.empty_like(r)
    vs1s = np.empty_like(r)
    vs2s = np.empty_like(r)
    betas = np.empty_like(r)
    Es = np.empty_like(r)
    for i, az in enumerate(azimuths):
        for j, phi in enumerate(zeniths):
            d = np.array(
                [np.cos(az) * np.sin(phi), np.sin(az) * np.sin(phi), -np.cos(phi)]
            )  # change_hemispheres
            velocities = mineral.wave_velocities(d)
            betas[i][j] = mineral.isentropic_linear_compressibility(d)
            Es[i][j] = mineral.isentropic_youngs_modulus(d)
            vps[i][j] = velocities[0][0]
            vs1s[i][j] = velocities[0][1]
            vs2s[i][j] = velocities[0][2]

    prps = []
    for type in plot_types:
        if type == "vp":
            prps.append(("V$_{P}$ (km/s)", vps / 1000.0))
        elif type == "vs1":
            prps.append(("V$_{S1}$ (km/s)", vs1s / 1000.0))
        elif type == "vs2":
            prps.append(("V$_{S2}$ (km/s)", vs2s / 1000.0))
        elif type == "vp/vs1":
            prps.append(("V$_{P}$/V$_{S1}$", vps / vs1s))
        elif type == "vp/vs2":
            prps.append(("V$_{P}$/V$_{S2}$", vps / vs2s))
        elif type == "s anisotropy":
            prps.append(
                ("S-wave anisotropy (%)", 200.0 * (vs1s - vs2s) / (vs1s + vs2s))
            )
        elif type == "linear compressibility":
            prps.append(("Linear compressibility (GPa$^{-1}$)", betas * 1.0e9))
        elif type == "youngs modulus":
            prps.append(("Youngs Modulus (GPa)", Es / 1.0e9))
        else:
            raise Exception("plot_type not recognised.")

    contour_sets = []
    ticks = []
    lines = []
    for i, prp in enumerate(prps):
        title, item = prp

        axes[i].set_title(title)

        vmin = np.min(item)
        vmax = np.max(item)
        spacing = np.power(10.0, np.floor(np.log10(vmax - vmin)))
        nt = int((vmax - vmin - vmax % spacing + vmin % spacing) / spacing)
        if nt == 1:
            spacing = spacing / 4.0
        elif nt < 4:
            spacing = spacing / 2.0
        elif nt > 8:
            spacing = spacing * 2.0

        tmin = vmin + (spacing - vmin % spacing)
        tmax = vmax - vmax % spacing
        nt = int((tmax - tmin) / spacing + 1)

        ticks.append(np.linspace(tmin, tmax, nt))
        contour_sets.append(
            axes[i].contourf(
                theta, r, item, n_divs, cmap=plt.cm.jet_r, vmin=vmin, vmax=vmax
            )
        )
        lines.append(
            axes[i].contour(
                theta, r, item, ticks[-1], colors=("black",), linewidths=(1,)
            )
        )
        axes[i].set_yticks([100])

    return contour_sets, ticks, lines



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.


from __future__ import absolute_import

import numpy as np
from sympy import Matrix, nsimplify
from collections import OrderedDict

from .material import material_property, cached_property
from .mineral import Mineral
from .solutionmodel import MechanicalSolution
from .solutionmodel import PolynomialSolution
from .averaging_schemes import reuss_average_function

from ..utils.reductions import independent_row_indices
from ..utils.chemistry import sum_formulae, sort_element_list_to_IUPAC_order


class Solution(Mineral):
    """
    This is the base class for all solutions.
    Site occupancies, endmember activities and the constant
    and pressure and temperature dependencies of the excess
    properties can be queried after using set_composition().
    States of the solution can only be queried after setting
    the pressure, temperature and composition using set_state().

    This class is available as :class:`burnman.Solution`.
    It uses an instance of :class:`burnman.SolutionModel` to
    calculate interaction terms between endmembers.

    All the solution parameters are expected to be in SI units.  This
    means that the interaction parameters should be in J/mol, with the T
    and P derivatives in J/K/mol and m^3/mol.

    The parameters are relevant to all solution models. Please
    see the documentation for individual models for details about
    other parameters.

    :param name: Name of the solution.
    :type name: string
    :param solution_model: The SolutionModel object defining the properties
        of the solution.
    :type solution_model: :class:`burnman.SolutionModel`
    :param molar_fractions: The molar fractions of each endmember in the solution.
        Can be reset using the set_composition() method.
    :type molar_fractions: numpy.array
    """

    def __init__(self, name=None, solution_model=None, molar_fractions=None):
        """
        Set up matrices to speed up calculations for when P, T, X is defined.
        """
        Mineral.__init__(self)

        # Solution needs a method attribute to call Mineral.set_state().
        # Note that set_method() below will not change self.method
        self.method = "SolutionMethod"

        if name is not None:
            self.name = name
        if solution_model is not None:
            self.solution_model = solution_model

        # Equation of state
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_method(
                self.solution_model.endmembers[i][0].params["equation_of_state"]
            )

        # Molar fractions
        if molar_fractions is not None:
            self.set_composition(molar_fractions)

    @cached_property
    def endmembers(self):
        return self.solution_model.endmembers

    def set_composition(self, molar_fractions):
        """
        Set the composition for this solution.
        Resets cached properties.

        :param molar_fractions: Molar abundance for each endmember, needs to sum to one.
        :type molar_fractions: list of float
        """
        assert len(self.solution_model.endmembers) == len(molar_fractions)

        if type(self.solution_model) != MechanicalSolution:
            assert sum(molar_fractions) > 0.9999
            assert sum(molar_fractions) < 1.0001

        if type(self.solution_model) == PolynomialSolution:
            self.solution_model.set_composition(molar_fractions)

        self.reset()
        self.molar_fractions = np.array(molar_fractions)

    def set_method(self, method):
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_method(method)
        # note: do not set self.method here!
        self.reset()

    def set_state(self, pressure, temperature):
        if type(self.solution_model) == PolynomialSolution:
            self.solution_model.set_state(pressure, temperature)

        Mineral.set_state(self, pressure, temperature)
        for i in range(self.n_endmembers):
            self.solution_model.endmembers[i][0].set_state(pressure, temperature)

    @material_property
    def formula(self):
        """
        Returns molar chemical formula of the solution.
        :rtype: Counter
        """
        return sum_formulae(self.endmember_formulae, self.molar_fractions)

    @material_property
    def site_occupancies(self):
        """
        :returns: The fractional occupancies of species on each site.
        :rtype: list of OrderedDicts
        """
        occs = np.einsum(
            "ij, i", self.solution_model.endmember_occupancies, self.molar_fractions
        )
        site_occs = []
        k = 0
        for i in range(self.solution_model.n_sites):
            site_occs.append(OrderedDict())
            for j in range(len(self.solution_model.sites[i])):
                site_occs[-1][self.solution_model.sites[i][j]] = occs[k]
                k += 1

        return site_occs

    def site_formula(self, precision=2):
        """
        Returns the molar chemical formula of the solution with site occupancies.
            For example, [Mg0.4Fe0.6]2SiO4.

        :param precision: Precision with which to print the site occupancies
        :type precision: int

        :returns: Molar chemical formula of the solution with site occupancies
        :rtype: str
        """
        split_empty = self.solution_model.empty_formula.split("[")
        formula = split_empty[0]
        for i, site_occs in enumerate(self.site_occupancies):
            formula += "["
            for species, occ in site_occs.items():
                formula += f"{species}{occ:0.{precision}f}"
            formula += split_empty[i + 1]
        return formula

    @material_property
    def activities(self):
        """
        Returns a list of endmember activities [unitless].
        """
        return self.solution_model.activities(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def activity_coefficients(self):
        """
        Returns a list of endmember activity coefficients
        (gamma = activity / ideal activity) [unitless].
        """
        return self.solution_model.activity_coefficients(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_internal_energy(self):
        """
        Returns molar internal energy of the mineral [J/mol].
        Aliased with self.energy
        """
        return self.molar_helmholtz + self.temperature * self.molar_entropy

    @material_property
    def excess_partial_gibbs(self):
        """
        Returns excess partial molar gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_gibbs_free_energies(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def excess_partial_volumes(self):
        """
        Returns excess partial volumes [m^3].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_volumes(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def excess_partial_entropies(self):
        """
        Returns excess partial entropies [J/K].
        Property specific to solutions.
        """
        return self.solution_model.excess_partial_entropies(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def partial_gibbs(self):
        """
        Returns endmember partial molar gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].gibbs
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_partial_gibbs
        )

    @material_property
    def partial_volumes(self):
        """
        Returns endmember partial volumes [m^3].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].molar_volume
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_partial_volumes
        )

    @material_property
    def partial_entropies(self):
        """
        Returns endmember partial entropies [J/K].
        Property specific to solutions.
        """
        return (
            np.array(
                [
                    self.solution_model.endmembers[i][0].molar_entropy
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_partial_entropies
        )

    @material_property
    def excess_gibbs(self):
        """
        Returns molar excess gibbs free energy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_gibbs_free_energy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def gibbs_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the Gibbs free energy [J]. Property specific to solutions.
        """
        return self.solution_model.gibbs_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def entropy_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the entropy [J/K]. Property specific to solutions.
        """
        return self.solution_model.entropy_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def volume_hessian(self):
        """
        Returns an array containing the second compositional derivative
        of the volume [m^3]. Property specific to solutions.
        """
        return self.solution_model.volume_hessian(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_gibbs(self):
        """
        Returns molar Gibbs free energy of the solution [J/mol].
        Aliased with self.gibbs.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].gibbs * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_gibbs
        )

    @material_property
    def molar_helmholtz(self):
        """
        Returns molar Helmholtz free energy of the solution [J/mol].
        Aliased with self.helmholtz.
        """
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the solution [kg/mol].
        """
        return sum(
            [
                self.solution_model.endmembers[i][0].molar_mass
                * self.molar_fractions[i]
                for i in range(self.n_endmembers)
            ]
        )

    @material_property
    def excess_volume(self):
        """
        Returns excess molar volume of the solution [m^3/mol].
        Specific property for solutions.
        """
        return self.solution_model.excess_volume(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the solution [m^3/mol].
        Aliased with self.V.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].molar_volume
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_volume
        )

    @material_property
    def density(self):
        """
        Returns density of the solution [kg/m^3].
        Aliased with self.rho.
        """
        return self.molar_mass / self.molar_volume

    @material_property
    def excess_entropy(self):
        """
        Returns excess molar entropy [J/K/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_entropy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the solution [J/K/mol].
        Aliased with self.S.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].S * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_entropy
        )

    @material_property
    def excess_enthalpy(self):
        """
        Returns excess molar enthalpy [J/mol].
        Property specific to solutions.
        """
        return self.solution_model.excess_enthalpy(
            self.pressure, self.temperature, self.molar_fractions
        )

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the solution [J/mol].
        Aliased with self.H.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].H * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.excess_enthalpy
        )

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the solution [Pa].
        Aliased with self.K_T.
        """
        return (
            self.V
            * 1.0
            / (
                sum(
                    [
                        self.solution_model.endmembers[i][0].V
                        / (self.solution_model.endmembers[i][0].K_T)
                        * self.molar_fractions[i]
                        for i in range(self.n_endmembers)
                    ]
                )
                + self.solution_model.VoverKT_excess()
            )
        )

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns adiabatic bulk modulus of the solution [Pa].
        Aliased with self.K_S.
        """
        if self.temperature < 1e-10:
            return self.isothermal_bulk_modulus
        else:
            return (
                self.isothermal_bulk_modulus
                * self.molar_heat_capacity_p
                / self.molar_heat_capacity_v
            )

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the solution.
        (or inverse isothermal bulk modulus) [1/Pa].
        Aliased with self.K_T.
        """
        return 1.0 / self.isothermal_bulk_modulus

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibility of the solution.
        (or inverse adiabatic bulk modulus) [1/Pa].
        Aliased with self.K_S.
        """
        return 1.0 / self.adiabatic_bulk_modulus

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the solution [Pa].
        Aliased with self.G.
        """
        G_list = np.fromiter(
            (e[0].G for e in self.solution_model.endmembers),
            dtype=float,
            count=self.n_endmembers,
        )
        return reuss_average_function(self.molar_fractions, G_list)

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the solution [m/s].
        Aliased with self.v_p.
        """
        return np.sqrt(
            (self.adiabatic_bulk_modulus + 4.0 / 3.0 * self.shear_modulus)
            / self.density
        )

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the solution [m/s].
        Aliased with self.v_phi.
        """
        return np.sqrt(self.adiabatic_bulk_modulus / self.density)

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the solution [m/s].
        Aliased with self.v_s.
        """
        return np.sqrt(self.shear_modulus / self.density)

    @material_property
    def grueneisen_parameter(self):
        """
        Returns grueneisen parameter of the solution [unitless].
        Aliased with self.gr.
        """
        if self.temperature < 1e-10:
            return float("nan")
        else:
            return (
                self.thermal_expansivity
                * self.isothermal_bulk_modulus
                * self.molar_volume
                / self.molar_heat_capacity_v
            )

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient (alpha)
        of the solution [1/K].
        Aliased with self.alpha.
        """
        return (1.0 / self.V) * (
            sum(
                [
                    self.solution_model.endmembers[i][0].alpha
                    * self.solution_model.endmembers[i][0].V
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.solution_model.alphaV_excess()
        )

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the
        solution [J/K/mol].
        Aliased with self.C_v.
        """
        return (
            self.molar_heat_capacity_p
            - self.molar_volume
            * self.temperature
            * self.thermal_expansivity
            * self.thermal_expansivity
            * self.isothermal_bulk_modulus
        )

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure
        of the solution [J/K/mol].
        Aliased with self.C_p.
        """
        return (
            sum(
                [
                    self.solution_model.endmembers[i][0].molar_heat_capacity_p
                    * self.molar_fractions[i]
                    for i in range(self.n_endmembers)
                ]
            )
            + self.solution_model.Cp_excess()
        )

    @cached_property
    def stoichiometric_matrix(self):
        """
        A sympy Matrix where each element M[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """

        def f(i, j):
            e = self.elements[j]
            if e in self.endmember_formulae[i]:
                return nsimplify(self.endmember_formulae[i][e])
            else:
                return 0

        return Matrix(len(self.endmember_formulae), len(self.elements), f)

    @cached_property
    def stoichiometric_array(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of atoms of element[j] in endmember[i].
        """
        return np.array(self.stoichiometric_matrix)

    @cached_property
    def reaction_basis(self):
        """
        An array where each element arr[i,j] corresponds
        to the number of moles of endmember[j] involved in reaction[i].
        """
        reaction_basis = np.array(
            [v[:] for v in self.stoichiometric_matrix.T.nullspace()]
        )

        if len(reaction_basis) == 0:
            reaction_basis = np.empty((0, len(self.endmember_names)))

        return reaction_basis

    @cached_property
    def n_reactions(self):
        """
        The number of reactions in reaction_basis.
        """
        return len(self.reaction_basis[:, 0])

    @cached_property
    def independent_element_indices(self):
        """
        A list of an independent set of element indices. If the amounts of
        these elements are known (element_amounts),
        the amounts of the other elements can be inferred by
        -compositional_null_basis[independent_element_indices].dot(element_amounts).
        """
        return sorted(independent_row_indices(self.stoichiometric_matrix.T))

    @cached_property
    def dependent_element_indices(self):
        """
        The element indices not included in the independent list.
        """
        return [
            i
            for i in range(len(self.elements))
            if i not in self.independent_element_indices
        ]

    @cached_property
    def compositional_null_basis(self):
        """
        An array N such that N.b = 0 for all bulk compositions that can
        be produced with a linear sum of the endmembers in the solution.
        """
        null_basis = np.array([v[:] for v in self.stoichiometric_matrix.nullspace()])

        M = null_basis[:, self.dependent_element_indices]
        assert (M.shape[0] == M.shape[1]) and (M == np.eye(M.shape[0])).all()

        return null_basis

    @cached_property
    def endmember_formulae(self):
        """
        A list of formulae for all the endmember in the solution.
        """
        return [mbr[0].params["formula"] for mbr in self.solution_model.endmembers]

    @cached_property
    def endmember_names(self):
        """
        A list of names for all the endmember in the solution.
        """
        return [mbr[0].name for mbr in self.solution_model.endmembers]

    @cached_property
    def n_endmembers(self):
        """
        The number of endmembers in the solution.
        """
        return len(self.solution_model.endmembers)

    @cached_property
    def elements(self):
        """
        A list of the elements which could be contained in the solution,
        returned in the IUPAC element order.
        """
        keys = []
        for f in self.endmember_formulae:
            keys.extend(f.keys())

        return sort_element_list_to_IUPAC_order(set(keys))


SolidSolution = Solution



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2022 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
Decker_1971
^^^^^^^^^^^
"""

import numpy as np
from ..eos.birch_murnaghan import birch_murnaghan
from ..eos import MGD2
from ..eos import debye
from ..classes.calibrant import Calibrant


class NaCl_B1(Calibrant):
    """
    The NaCl (B1 structured) pressure standard reported by Decker (1971).

    .. note:: This calibrant is not exactly the same as that proposed by Decker.
        The cold compression curve has here been approximated by a
        Birch-Murnaghan EoS.

    .. todo:: Make the calibrant exactly match that published by Decker.
    """

    def __init__(self):
        def _pressure_Decker_NaCl(volume, temperature, params):
            p300 = birch_murnaghan(params["V_0"] / volume, params)
            grueneisen = MGD2._grueneisen_parameter(0.0, params["V_0"] / volume, params)
            Debye_T = params["Debye_0"] * np.exp(
                (params["grueneisen_0"] - grueneisen) / params["q_0"]
            )
            Eqh = debye.thermal_energy(temperature, Debye_T, params["n"])
            EqhR = debye.thermal_energy(params["refT"], Debye_T, params["n"])
            dpqh = (Eqh - EqhR) * (grueneisen / volume)
            pressure = p300 + dpqh
            return pressure

        _params_Decker_NaCl = {
            "V_0": 2.7015e-05,
            "K_0": 24.02e9,
            "Kprime_0": 4.7369,
            "Debye_0": 279,
            "grueneisen_0": 1.59,
            "q_0": 0.93,
            "n": 2.0,
            "refT": 298.15,
            "P_0": 0,
            "Z": 4.0,
        }

        Calibrant.__init__(self, _pressure_Decker_NaCl, "pressure", _params_Decker_NaCl)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from sympy import Matrix
from scipy.linalg import block_diag

import logging
import importlib
from ..classes.polytope import MaterialPolytope, independent_row_indices
from ..classes.solution import Solution
from ..classes.composite import Composite
from .solution import transform_solution_to_new_basis

try:
    cp = importlib.import_module("cvxpy")
except ImportError as err:
    print(
        f"Warning: {err}. " "For full functionality of BurnMan, please install cvxpy."
    )


def solution_polytope_from_charge_balance(
    charges, charge_total, return_fractions=False
):
    """
    Creates a polytope object from a list of the charges for each species on
    each site and the total charge for all site-species.

    :param charges: 2D list containing the total charge for species j on site i,
        including the site multiplicity. So, for example,
        a solution with the site formula [Mg,Fe]3[Mg,Al,Si]2Si3O12 would
        have the following list: [[6., 6.], [4., 6., 8.]].
    :type charges: 2D list of floats
    :param charge_total: The total charge for all site-species per formula unit.
        The example given above would have charge_total = 12.
    :type charge_total: float
    :param return_fractions: Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
        Default is False.
    :type return_fractions: bool

    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    n_sites = len(charges)
    all_charges = np.concatenate(charges)
    n_site_elements = len(all_charges)
    equalities = np.empty((n_sites + 1, n_site_elements + 1))
    equalities[:-1, 0] = -1
    i = 0
    for i_site, site_charges in enumerate(charges):
        equalities[i_site, 1:] = [
            1 if (j >= i and j < i + len(site_charges)) else 0
            for j in range(n_site_elements)
        ]
        i += len(site_charges)

    equalities[-1, 0] = -charge_total
    equalities[-1, 1:] = all_charges

    pos_constraints = np.concatenate(
        (np.zeros((len(equalities[0]) - 1, 1)), np.identity(len(equalities[0]) - 1)),
        axis=1,
    )
    return MaterialPolytope(
        equalities, pos_constraints, return_fractions=return_fractions
    )


def solution_polytope_from_endmember_occupancies(
    endmember_occupancies, return_fractions=False
):
    """
    Creates a polytope object from a list of independent endmember occupancies.

    :param endmember_occupancies: 2D list containing the
        site-species occupancies j for endmember i.
        So, for example, a solution with independent endmembers
        [Mg]3[Al]2Si3O12, [Mg]3[Mg0.5Si0.5]2Si3O12, [Fe]3[Al]2Si3O12
        might have the following array:
        [[1., 0., 1., 0., 0.],
        [1., 0., 0., 0.5, 0.5],
        [0., 1., 1., 0., 0.]],
        where the order of site-species is Mg_A, Fe_A, Al_B, Mg_B, Si_B.
    :type endmember_occupancies: 2D numpy array

    :param return_fractions: Determines whether the created polytope object
        returns its attributes (such as endmember occupancies) as fractions
        or as floats.
    :type return_fractions: bool

    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    n_sites = sum(endmember_occupancies[0])
    n_occs = endmember_occupancies.shape[1]

    nullspace = np.array(Matrix(endmember_occupancies).nullspace(), dtype=float)

    equalities = np.zeros((len(nullspace) + 1, n_occs + 1))
    equalities[0, 0] = -n_sites
    equalities[0, 1:] = 1

    if len(nullspace) > 0:
        try:
            equalities[1:, 1:] = nullspace
        except ValueError:
            equalities[1:, 1:] = nullspace[:, :, 0]

    pos_constraints = np.concatenate(
        (np.zeros((len(equalities[0]) - 1, 1)), np.identity(len(equalities[0]) - 1)),
        axis=1,
    )

    return MaterialPolytope(
        equalities,
        pos_constraints,
        return_fractions=return_fractions,
        independent_endmember_occupancies=endmember_occupancies,
    )


def composite_polytope_at_constrained_composition(
    composite, composition, return_fractions=False
):
    """
    Creates a polytope object from a Composite object and a composition.
    This polytope describes the complete set of valid composite
    endmember amounts that satisfy the compositional constraints.

    :param composite: A composite containing one or more Solution and Mineral objects.
    :type composite: :class:`burnman.Composite` object

    :param composition: A dictionary containing the amounts of each element.
    :type composition: dict

    :param return_fractions: Determines whether the created polytope object returns its
        attributes (such as endmember occupancies) as fractions or as floats.
    :type return_fractions: bool


    :returns: A polytope object corresponding to the parameters provided.
    :rtype: :class:`burnman.polytope.MaterialPolytope` object
    """
    c_array = np.empty((composite.n_elements, 1))
    c_array[:, 0] = [
        -composition[e] if e in composition else 0.0 for e in composite.elements
    ]

    equalities = np.concatenate((c_array, composite.stoichiometric_array.T), axis=1)

    eoccs = []
    for i, ph in enumerate(composite.phases):
        if isinstance(ph, Solution):
            eoccs.append(ph.solution_model.endmember_occupancies.T)
        else:
            eoccs.append(np.ones((1, 1)))

    eoccs = block_diag(*eoccs)
    inequalities = np.concatenate((np.zeros((len(eoccs), 1)), eoccs), axis=1)

    return MaterialPolytope(
        equalities, inequalities, number_type="float", return_fractions=return_fractions
    )


def simplify_composite_with_composition(composite, composition):
    """
    Takes a composite and a composition, and returns the simplest composite
    object that spans the solution space at the given composition.

    For example, if the composition is given as {'Mg': 2., 'Si': 1.5, 'O': 5.},
    and the composite is given as a mix of Mg,Fe olivine and pyroxene
    solutions, this function will return a composite that only contains
    the Mg-bearing endmembers.

    :param composite: The initial Composite object.
    :type composite: :class:`burnman.Composite` object

    :param composition: A dictionary containing the amounts of each element.
    :type composition: dict

    :returns: The simplified Composite object
    :rtype: :class:`burnman.Composite` object
    """
    polytope = composite_polytope_at_constrained_composition(
        composite, composition, return_fractions=True
    )

    composite_changed = False
    new_phases = []
    mbr_amounts = polytope.endmember_occupancies
    i = 0
    for i_ph, n_mbrs in enumerate(composite.endmembers_per_phase):
        ph = composite.phases[i_ph]

        amounts = mbr_amounts[:, i : i + n_mbrs].astype(float)
        i += n_mbrs

        rank = np.linalg.matrix_rank(amounts, tol=1.0e-8)

        if rank < n_mbrs:
            if isinstance(ph, Solution) and rank > 0:
                if len(amounts) > 1:
                    c_mean = np.mean(amounts, axis=0)
                else:
                    c_mean = amounts[0]

                poly = solution_polytope_from_endmember_occupancies(
                    ph.solution_model.endmember_occupancies
                )
                dmbrs = poly.endmembers_as_independent_endmember_amounts

                x = cp.Variable(dmbrs.shape[0])
                objective = cp.Minimize(cp.sum_squares(x))
                constraints = [dmbrs.T @ x == c_mean, x >= 0]

                prob = cp.Problem(objective, constraints)
                prob.solve()

                mbr_indices = np.argsort(x.value)[::-1]
                ind_indices = [i for i in mbr_indices if x.value[i] > 1.0e-6]
                new_basis = dmbrs[ind_indices]

                # And now reduce the new basis if necessary
                new_basis = new_basis[independent_row_indices(new_basis)]

                if len(new_basis) < ph.n_endmembers:
                    logging.info(
                        f"Phase {i_ph} ({ph.name}) is "
                        "rank-deficient ({rank} < {n_mbrs}). "
                        "The transformed solution is described "
                        f"using {len(new_basis)} endmembers."
                    )

                    composite_changed = True
                    soln = transform_solution_to_new_basis(ph, new_basis)
                    new_phases.append(soln)
                else:
                    logging.info(
                        "This solution is rank-deficient "
                        f"({rank} < {n_mbrs}), "
                        "but its composition requires all "
                        "independent endmembers."
                    )
            else:
                composite_changed = True
                logging.info(
                    f"Phase {i_ph} ({ph.name}) removed from " "composite (rank = 0)."
                )
        else:
            new_phases.append(ph)

    if composite_changed:
        return Composite(new_phases)
    else:
        return composite
